from typing import Iterable

from sanic_macrobase.endpoint import SanicEndpoint
from sanic_macrobase.exceptions import RoutingErrorException


class Route(object):
    """
    Base route for HTTP sanic server
    """
    def __init__(self, handler: SanicEndpoint, uri, methods=frozenset({'GET'}), host=None,
                  strict_slashes=None, version=None, name=None):
        super(Route, self).__init__()

        if not isinstance(handler, SanicEndpoint):
            raise RoutingErrorException('Handler must be instance of SanicEndpoint class')

        self._handler = handler
        self._uri = uri
        self._methods = methods
        self._host = host
        self._strict_slashes = strict_slashes
        self._version = version

        self._name = name
        if self._name is None:
            self._name = self._handler.__name__

    @property
    def handler(self) -> SanicEndpoint:
        return self._handler

    @property
    def uri(self) -> str:
        return self._uri

    @property
    def methods(self) -> Iterable:
        return self._methods

    @property
    def host(self) -> str:
        return self._host

    @property
    def strict_slashes(self) -> bool:
        return self._strict_slashes

    @property
    def version(self) -> str:
        return self._version

    @property
    def name(self) -> str:
        return self._name



from typing import List, ClassVar
import logging.config
import asyncio
import uvloop

from . import helpers

from macrobase_driver.driver import MacrobaseDriver, Context
from macrobase_driver.driver import MacrobaseDriver, Context
from macrobase_driver.logging import get_logging_config
from macrobase_driver.hook import HookHandler

from sanic_macrobase.config import SanicDriverConfig
from sanic_macrobase.route import Route
from sanic_macrobase.hook import SanicHookNames

from structlog import get_logger
from sanic import Sanic, Blueprint
from sanic.config import Config


log = get_logger('SanicDriver')


class SanicDriver(MacrobaseDriver):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.name is None:
            self.name = 'SanicDriver'

        self.config = SanicDriverConfig()
        self._hooks: Dict[SanicHookNames, List[HookHandler]] = {}
        self._routes = []
        self._preload_server()

    def _preload_server(self):
        self._sanic = Sanic(name=self.name, log_config=get_logging_config(self.config))
        self._sanic.config = Config()

    def update_config(self, config: SanicDriverConfig):
        """
        Add sanic driver config
        """
        self.config.update(config)

    def add_hook(self, name: SanicHookNames, handler):
        if name not in self._hooks:
            self._hooks[name] = []

        self._hooks[name].append(HookHandler(self, handler))

    def add_routes(self, routes: List[Route]):
        """
        Add HTTP routes
        """
        self._routes.extend(routes)

    def _apply_routes(self):
        prefix = self.config.APP_BLUEPRINT

        if prefix is None or len(prefix) == 0:
            server = self._sanic
        else:
            server = Blueprint(prefix, url_prefix=prefix)

        if self.config.HEALTH_ENDPOINT:
            from .endpoint import HealthEndpoint
            server.add_route(HealthEndpoint(self.context, self.config), '/health', {'GET', 'POST'})

        [server.add_route(
            r.handler,
            r.uri,
            methods=r.methods,
            host=r.host,
            strict_slashes=r.strict_slashes,
            version=r.version,
            name=r.name) for r in self._routes]

        if isinstance(server, Blueprint):
            self._sanic.blueprint(server)

    def _apply_logging(self):
        self._logging_config = get_logging_config(self.config)
        logging.config.dictConfig(self._logging_config)

    def _apply_hooks(self):
        for name, handlers in self._hooks.items():
            for handler in handlers:
                self._sanic.listener(name.value)(handler)

        # async def lock_context(driver, context: Context, loop):
        #     context.lock()
        #
        # self._sanic.listener(SanicHookNames.after_server_start.value)(HookHandler(self, lock_context))
        pass

    def _prepare_server(self):
        self._sanic.config.from_object(self.config)
        # self._apply_logging()
        self._apply_hooks()
        self._apply_routes()

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

        self._prepare_server()

        try:
            self._sanic.run(
                host=self.config.APP_HOST,
                port=self.config.APP_PORT,
                debug=self.config.DEBUG,
                workers=self.config.WORKERS,
                access_log=self.config.ACCESS_LOG)
        except Exception as e:
            log.error(e)
            self._sanic.stop()



from enum import Enum

class SanicHookNames(Enum):
    before_server_start = 'before_server_start'
    after_server_start = 'after_server_start'
    before_server_stop = 'before_server_stop'
    after_server_stop = 'after_server_stop'



from http import HTTPStatus

from macrobase_driver.endpoint import Endpoint

from sanic.request import Request
from sanic.response import BaseHTTPResponse, text as TextResponse, json as JsonResponse, file as FileResponse, raw as RawResponse


class SanicEndpoint(Endpoint):

    @staticmethod
    def params_from_dictparams(params: dict):
        args = {}

        for key in params:
            value = params[key]

            if isinstance(value, list) and len(value) == 1:
                value = value[0]

            args[key] = value

        return args

    @staticmethod
    def import_body_match_info(request: Request) -> dict:
        if request.match_info is not None:
            return dict(request.match_info)

        return {}

    @staticmethod
    def import_body_json(request: Request) -> dict:
        if 'application/json' in request.content_type and request.json is not None:
            return dict(request.json)

        return {}

    @staticmethod
    def import_body_args(request: Request) -> dict:
        if request.method != 'POST' and request.args is not None and len(request.args) > 0:
            args = SanicEndpoint.params_from_dictparams(request.args)

            return dict(args)

        return {}

    @staticmethod
    def import_body_files(request: Request) -> dict:
        if request.files is not None and len(request.files) > 0:
            files = {}

            for key in request.files:
                if isinstance(request.files[key], list):
                    files[key] = [{'type': file.type, 'body': file.body, 'name': file.name} for file in
                                  request.files[key]]
                else:
                    files[key] = request.files

            return files

        return {}

    @staticmethod
    def import_body_form(request: Request) -> dict:
        if request.form is not None and len(request.form) > 0:
            args = request.form

            if 'multipart/form-data' in request.content_type:
                args = SanicEndpoint.params_from_dictparams(args)

            return dict(args)

        return {}

    @staticmethod
    def import_body_headers(request: Request) -> dict:
        headers = {}

        for header in request.headers:
            if header[:2].lower() == 'x-':
                headers[header] = request.headers[header]

        return headers

    @staticmethod
    async def make_response_raw(raw_str: str, code: int = 200) -> BaseHTTPResponse:
        return RawResponse(str.encode(raw_str), status=code)

    @staticmethod
    async def make_response_text(text: str, code: int = 200) -> BaseHTTPResponse:
        return TextResponse(text, status=code)

    @staticmethod
    async def make_response_json(code: int = 200, message: str = None, data: dict = None, error_code: int = None) -> BaseHTTPResponse:
        if data is not None:
            return JsonResponse(data)

        if message is None:
            message = HTTPStatus(code).phrase

        if error_code is None:
            error_code = code

        data = {
            'code': error_code,
            'message': message
        }

        return JsonResponse(data, status=code)

    @staticmethod
    async def make_response_file(filepath: str) -> BaseHTTPResponse:
        return await FileResponse(filepath)

    async def handle(self, request: Request, auth: dict = None, *args, **kwargs) -> BaseHTTPResponse:
        body = {}

        # aggregate data from all inputs
        body.update(self.import_body_match_info(request))
        body.update(self.import_body_json(request))
        body.update(self.import_body_args(request))
        body.update(self.import_body_files(request))
        body.update(self.import_body_form(request))
        body.update(self.import_body_headers(request))

        body['auth'] = auth

        return await self._method(request, body, *args, **kwargs)

    async def _method(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        method = request.method.lower()
        func_name = f'method_{method}'

        if hasattr(self, func_name):
            func = getattr(self, func_name)

            return await func(request, body, *args, **kwargs)
        else:
            return await self.make_response_json(code=405, message='Method Not Allowed')

    async def method_get(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_json(code=500, message=f'{request.method} Not Impl')

    async def method_head(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_json(code=500, message=f'{request.method} Not Impl')

    async def method_post(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_json(code=500, message=f'{request.method} Not Impl')

    async def method_put(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_json(code=500, message=f'{request.method} Not Impl')

    async def method_delete(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_json(code=500, message=f'{request.method} Not Impl')

    async def method_connect(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_json(code=500, message=f'{request.method} Not Impl')

    async def method_options(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_json(code=500, message=f'{request.method} Not Impl')

    async def method_trace(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_json(code=500, message=f'{request.method} Not Impl')

    async def method_patch(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_json(code=500, message=f'{request.method} Not Impl')

class HealthEndpoint(SanicEndpoint):

    async def method_get(self, request: Request, body: dict, *args, **kwargs) -> BaseHTTPResponse:
        return await self.make_response_raw('Health')



from macrobase_driver.hook import HookNames


def sanic_listner_from_hook(hook_name: HookNames) -> str:
    if hook_name.value == HookNames.before_server_start.value:
        return 'before_server_start'
    elif hook_name.value == HookNames.after_server_stop.value:
        return 'after_server_stop'

    return None


from sanic.exceptions import SanicException


class RoutingErrorException(SanicException):

    def __init__(self, message, *args, **kwargs):
        super().__init__(message, status_code=500, *args, **kwargs)



from setuptools import setup, find_packages

setup(
    name='sanic_macrobase',
    version='0.0.13',
    packages=find_packages(),
    url='https://github.com/mbcores/sanic-macrobase',
    license='MIT',
    author='Alexey Shagaleev',
    author_email='alexey.shagaleev@yandex.ru',
    description='Sanic driver for macrobase framework',
    install_requires=[
        'macrobase-driver>=0.0.14',
        'sanic==18.12.0',
        'structlog==19.1.0'
    ]
)



from macrobase_driver.config import DriverConfig


class SanicDriverConfig(DriverConfig):

    LOGO: str = """
 _____       _
|  __ \     (_)               
| |  | |_ __ ___   _____ _ __ 
| |  | | '__| \ \ / / _ \ '__|
| |__| | |  | |\ V /  __/ |   
|_____/|_|  |_| \_/ \___|_|sanic
"""

    DEBUG: bool = False
    WORKERS: int = 1
    ACCESS_LOG: bool = True
    HEALTH_ENDPOINT: bool = True

    APP_HOST: str = '0.0.0.0'
    APP_PORT: int = 8000
    APP_BLUEPRINT: str = ''
